"""
Special additions to Marshmallow schemas for parsing and serializing
data from the Sparrow database.

Taken from https://gist.github.com/om-henners/97bc3a4c0b589b5184ba621fd22ca42e
"""
from marshmallow_sqlalchemy.fields import Related, Nested
from marshmallow.fields import Field, Raw, DateTime
from marshmallow.exceptions import ValidationError
from marshmallow.fields import UUID as _UUID
from marshmallow.utils import is_collection
from geoalchemy2.shape import from_shape, to_shape
from shapely.geometry import mapping, shape
from datetime import datetime

from .util import primary_key
from macrostrat.utils import get_logger

log = get_logger(__name__)


class UUID(_UUID):
    def _deserialize(self, value, attr, data, **kwargs):
        """We need to deserialize to a string to make SQLAlchemy happy"""
        return str(self._validated(value))


class JSON(Raw):
    pass


class Geometry(Field):
    """
    Field for for parsing and serializing a PostGIS geometry
    type to GeoJSON

    Use shapely and geoalchemy2 to serialize / deserialize a point
    Does make a big assumption about the data being spat back out as
    JSON, but what the hey.
    """

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return mapping(to_shape(value))

    def _deserialize(self, value, attr, data, **kwargs):
        if value is None:
            return None
        return from_shape(shape(value), srid=4326)


class PassThroughRelated(Related):
    def _deserialize(self, value, attr=None, data=None, **kwargs):
        # In cases where we provide a SQLAlchemy model directly, we just pass that along.
        if isinstance(value, self.related_model):
            return value
        return super()._deserialize(value, attr, data, **kwargs)


class Enum(PassThroughRelated):
    """
    Enums are represented by a `Related` field, but we want to potentially
    be able to revise/extend this later without breaking external APIs.
    """

    pass


class SmartNested(Nested, PassThroughRelated):
    """This field allows us to resolve relationships as either primary keys or nested models."""

    # https://github.com/marshmallow-code/marshmallow/blob/dev/src/marshmallow/fields.py
    # https://github.com/marshmallow-code/marshmallow-sqlalchemy/blob/dev/src/marshmallow_sqlalchemy/fields.py
    # TODO: better conformance of this field to Marshmallow's OpenAPI schema generation
    # https://apispec.readthedocs.io/en/latest/using_plugins.html
    def __init__(
        self, name, *, only=None, exclude=(), many=False, unknown=None, **field_kwargs
    ):
        Nested.__init__(
            self,
            name,
            only=only,
            exclude=exclude,
            many=many,
            unknown=unknown,
            **field_kwargs,
        )
        PassThroughRelated.__init__(self, **field_kwargs)
        self._many = many
        self._instances = set()
        self.allow_none = True

        # Pass through allowed_nests configuration to child schema
        self._copy_config("allowed_nests")
        self._copy_config("_show_audit_id")

    def _deserialize(self, value, attr=None, data=None, **kwargs):
        if isinstance(value, self.related_model):
            return value
        # Better error message for collections.
        if not is_collection(value) and self._many:
            raise ValidationError("Provided a single object for a collection field")
        if is_collection(value) and not self._many:
            raise ValidationError("Provided a collection for a instance field")

        return Nested._deserialize(self, value, attr, data, **kwargs)

    def _copy_config(self, key):
        """Copy configuration from root model"""
        if hasattr(self.root, key):
            setattr(self.schema, key, getattr(self.root, key))

    def _serialize_related_key(self, value):
        """Serialize the primary key for a related model. In the (common) special
        case of a 1-column primary key, return just the value of that column; otherwise,
        return a map of {column_name: value}"""
        key = primary_key(value)
        if len(key.keys()) == 1:
            return list(key.values())[0]
        return key

    def _serialize_instance(self, value):
        if value is None:
            return None
        self._instances.add(value)

        return self._serialize_related_key(value)

    def _should_nest(self):
        other_name = self.related_model.__table__.name
        _allowed = self.root.allowed_nests
        return other_name in _allowed or _allowed == "all"

    def _serialize(self, value, attr, obj):
        # Pass through allowed_nests configuration to child schema
        # Unclear why we have to do this again, but a test fails
        self._copy_config("allowed_nests")
        self._copy_config("_show_audit_id")

        if self._should_nest():
            # Serialize as nested
            return super(Nested, self)._serialize(value, attr, obj)
        # If we don't want to nest
        if self._many:
            return [self._serialize_instance(v) for v in value]
        return self._serialize_instance(value)


class DateTimeExt(DateTime):
    def _deserialize(self, value, attr=None, data=None, **kwargs):
        # Allow python datetime objects to be deserialized.
        if isinstance(value, datetime):
            return value

        # Try to parse YYYY-MM-DD HH:MM:SS strings.
        try:
            date = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
            if isinstance(date, datetime):
                return date
        except ValueError:
            pass

        # Try to parse YYYY-MM-DD strings.
        try:
            date = datetime.strptime(value, "%Y-%m-%d")
            if isinstance(date, datetime):
                return date
        except ValueError:
            pass

        return super()._deserialize(value, attr, data, **kwargs)



import marshmallow
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema
from marshmallow_sqlalchemy.fields import Related
from marshmallow.fields import Nested
from marshmallow_jsonschema import JSONSchema
from marshmallow_sqlalchemy.fields import get_primary_keys, ensure_list
from marshmallow.decorators import pre_load, post_load, post_dump
from sqlalchemy.orm import RelationshipProperty, joinedload, defer
from collections.abc import Mapping
from sqlalchemy import inspect

from .fields import SmartNested
from .util import is_pk_defined, pk_values, prop_is_required
from .converter import SparrowConverter, allow_nest, exclude_fields

from macrostrat.utils import get_logger

log = get_logger(__name__)


class SparrowSchemaError(Exception):
    ...


def _jsonschema_type_mapping(self):
    """TODO: this is just a shim"""
    return {"type": "integer"}


Related._jsonschema_type_mapping = _jsonschema_type_mapping
Nested._jsonschema_type_mapping = _jsonschema_type_mapping

json_schema = JSONSchema()


class BaseMeta:
    model_converter = SparrowConverter
    # Needed for SQLAlchemyAutoSchema
    include_relationships = True
    load_instance = True


def is_model_ready(model, data):
    mapper = inspect(model)
    required_props = [a for a in mapper.attrs if prop_is_required(a)]
    for attr in required_props:
        val = data.get(attr.key, None)
        if val is None:
            return False
    return True


def matches(instance, filters):
    for k, v in filters.items():
        try:
            if getattr(instance, k) != v:
                return False
        except AttributeError:
            return False
    return True


class ModelSchema(SQLAlchemyAutoSchema):
    """
    :param: allowed_nests: List of strings or "all"
    """

    # A list of SQLAlchemy collections that we are nesting on
    nest_collections = []

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("unknown", marshmallow.EXCLUDE)
        nests = kwargs.pop("allowed_nests", [])
        self.allowed_nests = nests
        model = self.opts.model.__name__
        _message = f"Setting up schema for model {model}"
        if len(self.allowed_nests) > 0:
            _message += f" with allowed nests: {nests}"
            log.debug(_message)

        # Excluding kinda works but still pulls back fields into schema
        self._deferred_fields = [
            *kwargs.get("exclude", []),
            # Global defer list (we should refactor this eventually)
            *exclude_fields.get(model, []),
        ]
        # log.debug(_message)
        self._show_audit_id = kwargs.pop("audit_id", False)
        self.__instance_cache = {}

        self._mapper = self.opts.model.__mapper__
        super().__init__(*args, **kwargs)

    @classmethod
    def query(cls, session):
        """A helper that allows query objects to be constructed:
        `UserSchema.query(db.session).all()`
        """
        return session.query(cls.opts.model)

    @classmethod
    def table_name(cls):
        return cls.opts.model.__table__.name

    @property
    def _table(self):
        return self.opts.model.__table__

    def _nested_relationships(self, *, nests=None, current_depth=0, max_depth=None):
        """Get relationship fields for nested models, allowing them
        to be pre-joined.

        kwargs:
        mode [nested, related, hybrid]
            - nested: loads only the models directly referenced in nests
            - related: loads all models, even those referenced in related models only
            - hybrid [default]: loads all models, attempting to pull back only needed
                attributes that define the relationship...
        """
        depth = current_depth
        if max_depth != None and depth > max_depth:
            return [], None
        # It would be nice if we didn't have to pass nests down here...
        _nests = nests or self.allowed_nests
        for key, field in self.fields.items():

            if not isinstance(field, Related):
                continue
            # Yield this relationship
            this_relationship = getattr(self.opts.model, key)

            nested = isinstance(field, SmartNested) and field._should_nest()

            ##load_hint needs to be the attribute of the model joining
            load_hint = None
            # We are passing through load_hints to allow us to
            # return only a subset of columns when primary keys are
            # desired.
            # if not nested:
            #    load_hint = this_relationship.property.remote_side

            yield [this_relationship], load_hint
            if not nested:
                continue
            # We're only working with nested fields now
            field.schema.allowed_nests = _nests
            # Get relationships from nested models
            for rel, load_hint in field.schema._nested_relationships(
                nests=_nests, current_depth=depth + 1, max_depth=max_depth
            ):
                yield [this_relationship, *rel], load_hint

    def query_options(self, max_depth=None):
        for rel, load_hint in self._nested_relationships(max_depth=max_depth):
            res = joinedload(*rel)
            if load_hint is not None:
                res = res.load_only(*list(load_hint))
            yield res
        # Defer excluded fields
        # Results in a huge speedup if fields are sizeable
        for field in self._deferred_fields:
            yield defer(field)

    def nested_relationships(self, max_depth=None):
        return [v for v, _ in self._nested_relationships(max_depth=max_depth)]

    def _build_filters(self, data):
        # Filter on properties that actually have a local column
        filters = {}

        # Try to get filters for unique and pk columns first
        for prop in self.opts.model.__mapper__.iterate_properties:
            columns = getattr(prop, "columns", False)
            if columns:
                is_fully_defined = all(
                    [any([c.primary_key, c.unique]) for c in columns]
                )
                # Shim for the fact that we don't correctly find Session.uuid as unique at the moment...
                # TODO: fix this in general
                # if self.opts.model.__name__ == "Session" and prop.key == "uuid":
                #    is_fully_defined = True
                if is_fully_defined:
                    val = data.get(prop.key, None)
                    if val is not None:
                        filters[prop.key] = val
        if filters:
            return filters

        for prop in self.opts.model.__mapper__.iterate_properties:
            val = data.get(prop.key, None)
            if getattr(prop, "uselist", False):
                continue

            if val is not None:
                if hasattr(prop, "direction"):
                    # For relationships
                    if is_pk_defined(val):
                        filters[prop.key] = val
                else:
                    filters[prop.key] = val
            elif hasattr(prop, "direction"):
                # This is unsatisfying, as we can't filter on pre-existing
                # related fields
                filters[prop.key] = None
        return filters

    def load(self, data, **kwargs):
        if self.session is None:
            return super().load(data, **kwargs)
        with self.session.no_autoflush:
            return super().load(data, **kwargs)

    def _get_instance(self, data):
        """Gets pre-existing instances if they are available."""

        if self.instance is not None:
            return self.instance

        if isinstance(data, self.opts.model):
            return data

        filters = self._build_filters(data)

        # Create "fast paths" to make sure we don't grab data and analysis
        # if they aren't properly linked. NOTE: we can probably solve this
        # in general by requiring nullable foreign keys to be specified
        if self.opts.model.__name__ == "datum":
            if filters.get("_analysis") is None:
                return None

        if self.opts.model.__name__ == "analysis":
            if filters.get("_session") is None:
                return None

        msg = f"Finding instance of {self.opts.model.__name__}"

        with self.session.no_autoflush:
            # Need to get relationship columns for primary keys!
            query = self.session.query(self.opts.model).filter_by(**filters)
            instance = query.first()

            if instance is not None:
                log.info(msg + f"...success!\n...filters: {filters}")
                for k, v in data.items():
                    setattr(instance, k, v)
                return instance
            else:
                log.info(msg + f"...none found\n...filters: {filters}")
            log.debug(data)
        return super().get_instance(data)

    @pre_load
    def expand_primary_keys(self, value, **kwargs):
        # If we have a database model, leave it alone.
        if isinstance(value, self.opts.model) or isinstance(value, Mapping):
            return value

        pk_vals = ensure_list(value)
        # log.debug("Expanding keys " + str(pk_vals))

        pk = get_primary_keys(self.opts.model)
        try:
            assert len(pk) == len(pk_vals)
        except AssertionError:
            raise SparrowSchemaError(
                f"Could not expand primary key for {self.opts.model.__name__} from {value}"
            )
        return {col.key: val for col, val in zip(pk, pk_vals)}

    def _get_cached_instance(self, data):
        """Get an instance cached within this load transaction"""
        cache_key = None
        try:
            cache_key = hash(frozenset(data.items()))
            match_ = self.__instance_cache.get(cache_key, None)
            if match_ is not None:
                log.debug(
                    f"Found {match_} in session cache for {data} (key: {cache_key})"
                )
                return match_, cache_key
        except TypeError:
            pass
        return None, cache_key

    @post_load
    def make_instance(self, data, **kwargs):
        # We start off with no instance loaded
        instance = None
        # FIRST, try to load from cached values created during this import
        #
        # This cache exists primarily to cover cases where we import models that need to be
        # linked in multiple places (e.g. units, parameters, and datum_types). We could potentially
        # expand it to cover cases where we want to create several models up front and then reference
        # them from other places in the import model hierarchy.
        #
        # Datum (and some other analytical models) can easily be pulled in without a specific provided
        # identity (they get their uniqueness in part from their link to analysis). Given this,
        # it is possible for us to "steal" them from other analyses imported
        # at the same time. It's likely this could happen with analysis and other models as well.
        #
        # See https://github.com/EarthCubeGeochron/Sparrow/issues/80
        #
        # We solve this by exempting these models from caching. However, there are
        # likely more issues that will arise along these lines in the near future.
        cache_key = None
        # We should figure out a better way to do this than a blacklist...or at
        # least pull the blacklist into config code somewhere.
        _cache_blacklist = ["datum", "analysis", "session", "sample"]
        if self.opts.model.__name__ not in _cache_blacklist:
            instance, cache_key = self._get_cached_instance(data)

        # THEN, try to load from existing instances
        if instance is None and not self.transient:
            instance = self._get_instance(data)

        # FINALLY, make a new instance if one can't be found
        if instance is None:
            instance = self.opts.model(**data)
            if self.session is not None:
                log.debug(f"Adding new {instance} to session")
                self.session.add(instance)
        if cache_key is not None:
            self.__instance_cache[cache_key] = instance
        return instance

    @post_dump
    def remove_internal_fields(self, data, many, **kwargs):
        if not self._show_audit_id:
            data.pop("pgmemento_audit_id", None)
        return data

    def to_json_schema(self, model):
        return json_schema.dump(model)

    def _nest_relationships(self):
        model = self.opts.model
        nests = []
        for prop in model.__mapper__.iterate_properties:
            if not isinstance(prop, RelationshipProperty):
                continue
            if allow_nest(model.__table__.name, prop.target.name):
                nests.append(prop)
        return nests

    def _available_nests(self):
        return [rel.target.name for rel in self._nest_relationships()]

    from .display import pretty_print



from marshmallow_sqlalchemy import exceptions

from .schema import ModelSchema, BaseMeta
from .util import to_schema_name

from macrostrat.database.mapper.utils import ModelCollection, classname_for_table
from macrostrat.utils import get_logger

log = get_logger(__name__)


def model_interface(model, session=None) -> ModelSchema:
    """
    Create a Marshmallow interface to a SQLAlchemy model
    """
    # Create a meta class
    metacls = type(
        "Meta",
        (BaseMeta,),
        dict(
            model=model,
            sqla_session=session,
            load_instance=False,
            include_relationships=True,
        ),
    )

    schema_name = to_schema_name(model.__name__)
    try:
        # All conversion logic comes from ModelSchema
        return type(schema_name, (ModelSchema,), {"Meta": metacls})
    except exceptions.ModelConversionError as err:
        log.error(type(err).__name__ + ": " + schema_name + " - " + str(err))
        return None


class InterfaceCollection(ModelCollection):
    def register(self, *classes):
        for cls in classes:
            # Don't build model for spatial_ref_sys PostGIS table
            # TODO: we could probably generalize this.
            if cls.__table__.name == "spatial_ref_sys":
                continue
            if cls.__table__.schema == "enum":
                continue
            self._register_model(cls)

    def _register_model(self, cls):
        k = classname_for_table(cls.__table__)
        # Bail if we have a view
        if not hasattr(cls, "__mapper__"):
            return
        self.add(k, model_interface(cls))



from marshmallow_sqlalchemy import ModelConverter

# We need to use marshmallow_sqlalchemy's own implementation
# of the 'Nested' field in order to pass along the SQLAlchemy session
# to nested schemas.
# See https://github.com/marshmallow-code/marshmallow-sqlalchemy/issues/67
from marshmallow_sqlalchemy.fields import Related, RelatedList

import geoalchemy2 as geo
from sqlalchemy.orm import RelationshipProperty
from sqlalchemy.types import Integer, Numeric, DateTime
from sqlalchemy.dialects import postgresql
from macrostrat.utils import get_logger

from macrostrat.database.mapper.utils import trim_postfix
from .fields import (
    Geometry,
    Enum,
    JSON,
    SmartNested,
    UUID,
    DateTimeExt,
    PassThroughRelated,
)
from .util import to_schema_name


log = get_logger(__name__)

# Control how relationships can be resolved for import and serialization
# TODO:
# - we might want to slightly decouple the import and representation sides of this
# - we also might want to move this to a configuration section so it can be modified
allowed_collections = {
    "data_file": ["data_file_link"],
    "data_file_link": ["session", "sample", "analysis", "instrument_session"],
    "sample": [
        "session",
        "material",
        "project",
        "sample_geo_entity",
        "researcher",
        "publication",
        "tag",
        "attribute",
    ],
    "geo_entity": "all",
    "sample_geo_entity": "all",
    "session": [
        "analysis",
        "attribute",
        "project",
        "publication",
        "sample",
        "instrument",
        "instrument_session",
        "target",
        "method",
        "tag",
    ],
    "analysis": ["datum", "attribute", "constant", "analysis_type", "material", "tag"],
    "attribute": ["parameter", "unit"],
    "instrument_session": ["session", "project", "researcher"],
    "project": [
        "researcher",
        "publication",
        "session",
        "sample",
        "instrument_session",
        "tag",
    ],
    "datum": ["datum_type", "tag"],
    "datum_type": ["parameter", "unit", "error_unit", "error_metric"],
}

# This field exclusion only makes sense for LaserChron right now
exclude_fields = {"data_file": ["csv_data"]}


def allow_nest(outer, inner):
    if outer == inner:
        # Tables can always nest themselves
        return True
    coll = allowed_collections.get(outer, [])
    if coll == "all":
        return True
    return inner in coll


class SparrowConverter(ModelConverter):
    # Make sure that we can properly convert geometries
    # and geographies
    SQLA_TYPE_MAPPING = dict(
        list(ModelConverter.SQLA_TYPE_MAPPING.items())
        + list(
            {
                geo.Geometry: Geometry,
                geo.Geography: Geometry,
                postgresql.JSON: JSON,
                postgresql.JSONB: JSON,
                postgresql.UUID: UUID,
                DateTime: DateTimeExt,
            }.items()
        )
    )

    def _get_key(self, prop):
        # Only change columns for relationship properties
        if not isinstance(prop, RelationshipProperty):
            return prop.key

        # Relationship with a single local column should be named for its
        # column (less '_id' postfix)
        if hasattr(prop, "local_columns") and len(prop.local_columns) == 1:
            col_name = list(prop.local_columns)[0].name
            # Special case for 'data_file_link'
            if col_name == "file_hash":
                return prop.target.name
            if col_name != "id":
                return trim_postfix(col_name, "_id")

        # One-to-many models should have the field
        # named after the local column
        if prop.secondary is None and not prop.uselist:
            return prop.key

        # For tables not in Sparrow's standard schemas, we make sure to append the
        # schema name in front of keys, in order for views etc. to not trample
        # on already-used names. This was added in order to solve problems
        # with the automapping of `core_view.datum`.
        #
        # It may be desirable to improve this by adding a __prefix before views,
        # OR by making _ALL_ non-public schema tables require a prefix.
        if prop.target.schema not in [None, "vocabulary", "enum"]:
            return f"{prop.target.schema}_{prop.target.name}"

        # Otherwise, we go with the name of the target model.
        return str(prop.target.name)

    def fields_for_model(self, model, **kwargs):
        fields = super().fields_for_model(model, **kwargs)
        return {k: v for k, v in fields.items() if v is not None}

    def _should_exclude_field(self, prop, fields=None, exclude=None):
        if fields and prop.key not in fields:
            return True
        if exclude and prop.key in fields:
            return True

        # Get the name of this table
        this_table = prop.parent.tables[0]

        # Exclude fields referenced in our `exclude_fields` dict
        if prop.key in exclude_fields.get(this_table.name, []):
            return True

        if not isinstance(prop, RelationshipProperty):
            return False

        # ## Deal with relationships here #

        # if prop.target.name == "data_file_link":
        #     # Data files are currently exclusively dealt with internally...
        #     # (this may change)
        #     return True

        # # Exclude field based on table name
        other_table = prop.target

        if other_table == this_table and prop.uselist:
            # Don't allow self-referential collections
            return True

        if prop.uselist and not allow_nest(this_table.name, other_table.name):
            # Disallow list fields that aren't related (these usually don't have
            # corresponding local columns)
            return True

        # if this_table.name == "datum_type" and other_table.name in [
        #     "constant",
        #     "datum",
        # ]:
        # Fields that are not allowed to be nested
        #    return True

        return False

    def _get_field_kwargs_for_property(self, prop):
        # Somewhat ugly method, mostly to decide if field is dump_only or required
        kwargs = super()._get_field_kwargs_for_property(prop)
        kwargs["data_key"] = self._get_key(prop)

        if isinstance(prop, Numeric):
            kwargs["allow_nan"] = True

        if isinstance(prop, RelationshipProperty):  # Relationship property
            cols = list(getattr(prop, "local_columns", []))
            kwargs["required"] = False
            for col in cols:
                if not col.nullable:
                    kwargs["required"] = True
            if prop.uselist:
                kwargs["required"] = False
            # # Empty collections can be represented by null values
            # if not kwargs["required"]:
            #     kwargs["allow_none"] = True

            return kwargs

        for col in prop.columns:
            is_integer = isinstance(col.type, Integer)
            # Special case for audit columns
            if col.name == "pgmemento_audit_id" and is_integer:
                kwargs["dump_only"] = True
                return kwargs

            if is_integer and col.primary_key:
                # Integer primary keys should be dump-only, probably
                # We could probably check for a default sequence
                # if we wanted to be general...
                kwargs["dump_only"] = True
            elif not col.nullable:
                kwargs["required"] = True

            dump_only = kwargs.get("dump_only", False)
            if not dump_only and not col.nullable:
                kwargs["required"] = True
            if dump_only:
                kwargs["required"] = False
            if col.default is not None:
                kwargs["required"] = False

            # We allow setting of UUIDs for now, but maybe we shouldn't
            if isinstance(col.type, postgresql.UUID):
                # This should be covered by the "default" case, but for some reason
                # server-set defaults don't show up
                kwargs["required"] = False

        return kwargs

    def _related_entity(self, prop):
        # If this is a relationship, return related entity
        if isinstance(prop, RelationshipProperty):
            return prop.mapper.entity
        return None

    def property2field(self, prop, **kwargs):
        """This override improves our handling of relationships"""
        if isinstance(prop, RelationshipProperty):
            return self._property2relationship(prop, **kwargs)
        else:
            return super().property2field(prop, **kwargs)

    def _property2relationship(self, prop, **kwargs):
        # The "name" is actually the name of the related model, NOT the name
        # of field
        cls = self._related_entity(prop)
        name = to_schema_name(cls.__name__)

        field_kwargs = self._get_field_kwargs_for_property(prop)
        field_kwargs.update(**kwargs)

        this_table = prop.parent.tables[0]
        if not allow_nest(this_table.name, prop.target.name):
            # Fields that are not allowed to be nested
            if prop.uselist:
                return RelatedList(PassThroughRelated, **field_kwargs)
            else:
                return PassThroughRelated(**field_kwargs)
        if prop.target.schema == "enum":
            # special carve-out for enums represented as foreign keys
            # (these should be stored in the 'enum' schema):
            return Enum(**field_kwargs)

        # Ignore fields that reference parent models in a nesting relationship
        exclude = []
        for p in cls.__mapper__.relationships:
            if self._should_exclude_field(p):
                # Fields that are already excluded do not need to be excluded again.
                continue
            id_ = self._get_field_name(p)
            if p.mapper.entity == prop.parent.entity:
                exclude.append(id_)

        return SmartNested(name, many=prop.uselist, exclude=exclude, **field_kwargs)



#!/usr/bin/env python
from typing import Any
from os import environ
from pathlib import Path
from typing import Any
from contextvars import ContextVar


from macrostrat.database.mapper import DatabaseModelCache
from macrostrat.database.mapper.utils import (
    _classname_for_table,
    name_for_collection_relationship,
    name_for_scalar_relationship,
)
from .util import _gen_relationship

_model_cache: ContextVar[Any] = ContextVar("model-cache", default=None)

cache_file = Path(__file__).parent / "sparrow-db-models.pickle"

environ["SPARROW_CACHE_DATABASE_MODELS"] = "1"
environ["SPARROW_DATABASE_MODEL_CACHE_PATH"] = str(cache_file)

cacher = DatabaseModelCache(cache_file=str(cache_file))


def get_automap_base():
    cache = _model_cache.get()
    if cache is not None:
        return cache
    cacher._load_database_map()
    base = cacher.automap_base()

    kwargs = dict(
        name_for_scalar_relationship=name_for_scalar_relationship,
        name_for_collection_relationship=name_for_collection_relationship,
        classname_for_table=_classname_for_table,
        generate_relationship=_gen_relationship,
    )

    for schema in ["vocabulary", "core_view", "tags", "public"]:
        base.prepare(None, schema=schema, **kwargs)

    _model_cache.set(base)
    return base


def get_model(name):
    base = get_automap_base()
    return getattr(base.classes, name)



from marshmallow.fields import Nested
from marshmallow.exceptions import RegistryError
from click import echo, secho, style
from shutil import get_terminal_size
from .fields import PassThroughRelated
from .util import to_schema_name


def styled_key(k, **kwargs):
    return style("• ", dim=True) + style(k, **kwargs)


sep = "…"
indent = 3


class ModelPrinter(object):
    def __init__(self, nest_level=0, show_dump_only=False):
        self.nest_level = nest_level
        self.width, _ = get_terminal_size()
        self.show_dump_only = show_dump_only

    def print_nested(self, k, field, level=0):
        prefix = level * indent * " "
        try:
            if level < self.nest_level:
                self.print_model(
                    field.schema, level=level + 1, key=k, exclude=field.exclude
                )
            else:
                self.print_field(
                    k, field.schema, bold=True, level=level, required=field.required
                )
        except (ValueError, RegistryError, AttributeError) as err:
            echo(prefix + styled_key(k) + "  " + style(str(err), fg="red"))

    def print_field(self, key, field, level=0, **kwargs):
        prefix = level * indent * " "
        classname = field.__class__.__name__

        if isinstance(field, PassThroughRelated):
            name = to_schema_name(field.related_model.__name__)
            classname = name
            kwargs["bold"] = True
        if classname == "Geometry":
            classname = "Geometry (GeoJSON)"

        required = getattr(field, "required", kwargs.pop("required", False))

        many_to_one = getattr(field, "many", False)

        row = build_row_styles(
            key,
            classname,
            prefix=prefix,
            level=level,
            width=self.width,
            dump_only=field.dump_only,
            required=required,
            many_to_one=many_to_one,
            **kwargs,
        )

        echo(row)

    def __sort_fields(self, a):
        return isinstance(a[1], Nested)

    def print_model(self, model, key=None, level=0, exclude=[], model_alias=None):
        prefix = level * indent * " "

        new_prefix = prefix[0 : len(prefix) - indent]
        if key is not None:
            self.print_field(key, model, level=level - 1, bold=True)
        else:
            model_name = new_prefix + style(model.__class__.__name__, bold=True)
            echo(model_name, nl=False)
            if model_alias is not None:
                echo(style(" / ", dim=True) + str(model_alias))
            else:
                echo()

        fields = list(model.declared_fields.items())
        fields.sort(key=self.__sort_fields)

        for k, v in fields:
            name = v.data_key or k
            if v.dump_only and not self.show_dump_only:
                continue
            if k in exclude:
                continue
            if isinstance(v, Nested):
                self.print_nested(name, v, level=level)
            else:
                self.print_field(name, v, level=level)
        if len(exclude):
            excluded_names = [model.declared_fields[k].data_key or k for k in exclude]
            secho(prefix + "  excluded: " + " ".join(excluded_names), dim=True)

    def __call__(self, model, model_alias=None):
        self.print_model(model, model_alias=model_alias)


def pretty_print(model, nested=0, model_alias=None, show_dump_only=False):
    printer = ModelPrinter(nest_level=nested, show_dump_only=show_dump_only)
    printer(model, model_alias=model_alias)


def build_row_styles(
    key,
    classname,
    prefix="",
    level=0,
    width=80,
    dump_only=False,
    required=False,
    many_to_one=False,
    **kwargs,
):
    modifier = ""
    if many_to_one:
        modifier = "..."

    nfill = width - 4 - level * indent - len(key) - len(classname) - len(modifier)
    dim = dump_only

    return "".join(
        [
            prefix,
            styled_key(key, dim=dim, underline=required),
            style(modifier, dim=dim, bold=True),
            " ",
            style(sep * nfill, dim=True),
            " ",
            style(classname, fg="cyan", dim=True, **kwargs),
        ]
    )


def print_field(*args, **kwargs):
    kwargs["width"] = 40
    kwargs["prefix"] = 2 * indent * " "
    row = build_row_styles(*args, **kwargs)
    echo(row)


def print_key(show_dump_only=False):
    nfill = 40 - 13
    echo(2 * indent * " " + style("Key") + " " * nfill + style("Model type", dim=True))
    print_field("required", "String", required=True)
    print_field("optional", "Integer", required=False)
    print_field("nested", "AnotherModel", required=False, bold=True)
    print_field(
        "many-to-one", "AnotherModel", required=False, bold=True, many_to_one=True
    )
    if show_dump_only:
        print_field("dump-only", "String", required=False, dump_only=True)



from typing import List
from macrostrat.database.mapper.utils import ModelCollection
from marshmallow.utils import RAISE
from .interface import model_interface, InterfaceCollection
from .schema import ModelSchema, BaseMeta

from .schema_checking import get_automap_base, get_model
from .display import print_key


def get_cached_models():
    """Get a collection of all Sparrow database models available for importing"""
    base = get_automap_base()
    return ModelCollection(base.classes)


def get_all_loader_schemas():
    """Get a collection of all Sparrow database loader schemas"""
    models = get_cached_models()
    coll = InterfaceCollection(models)
    return coll


def get_loader_schema(name):
    """Get the loader schema for a Sparrow database model"""
    return getattr(get_all_loader_schemas(), name)


def validate_data(model_name, data):
    """Check a dictionary of data against a Sparrow database model"""

    interface = get_loader_schema(model_name)
    return interface(transient=True).load(data, unknown=RAISE)


def show_loader_schemas(*schemas: List[str], nest_depth=0, show_dump_only=False):
    """Print the loader schema for a Sparrow database model"""
    models = get_cached_models()
    coll = InterfaceCollection(models)

    if len(schemas) == 0:
        schemas = coll.keys()

    schemas = list(schemas)
    schemas.sort()
    prepended = [
        "project",
        "sample",
        "session",
        "analysis",
        "datum",
        "datum_type",
    ]
    schemas = [s for s in schemas if s not in prepended]
    schemas = prepended + schemas

    for name in schemas:
        _schema = getattr(coll, name)
        schema = _schema()
        schema.pretty_print(
            nested=nest_depth, model_alias=name, show_dump_only=show_dump_only
        )
        print()
    print_key(show_dump_only=show_dump_only)


def show_loader_schema(schema: str, nest_depth=0):
    """Print the loader schema for a Sparrow database model"""
    show_loader_schemas(schema, nest_depth=nest_depth)



import typer
import warnings
from rich import print
from typing import List
from marshmallow import ValidationError
from . import show_loader_schemas, validate_data
import json
import sys

app = typer.Typer()


@app.command(
    name="show-schemas",
    help="Show loader schemas for Sparrow database models",
)
def show_schemas(schemas: List[str] = [], nest_depth: int = 0):
    """Print the loader schema for a Sparrow database model"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        show_loader_schemas(*schemas, nest_depth=nest_depth)


@app.command(
    name="validate",
    help="Validate data piped from stdin",
)
def _validate_data(model_name: str):
    """Check a dictionary of data against a Sparrow database model"""
    data = json.load(sys.stdin)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        try:
            res = validate_data(model_name, data)
            print("[bold green]Success![/bold green]")
            print(res)
        except ValidationError as e:
            print("[bold red]Validation error![/bold red]")
            print(e)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sparrow', 'sparrow.loader']

package_data = \
{'': ['*']}

install_requires = \
['GeoAlchemy2>=0.9.4,<0.10.0',
 'SQLAlchemy>=1.4.41,<2.0.0',
 'macrostrat.database>=1.0.2,<2.0.0',
 'marshmallow-jsonschema>=0.10.0,<0.11.0',
 'marshmallow-sqlalchemy>=0.24.2,<0.25.0',
 'marshmallow>=3.11.1,<4.0.0',
 'shapely>=1,<2',
 'stringcase>=1.2.0,<2.0.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['sparrow-loader = sparrow.loader.cli:app']}

setup_kwargs = {
    'name': 'sparrow-loader',
    'version': '1.0.4',
    'description': 'Utilities for loading data into Sparrow',
    'long_description': "# sparrow.loader\n\nThe `sparrow.loader` module helps you prepare data for loading\ninto the Sparrow geochemistry database system.\n\nWhen disconnected from a database, it can be used to check that\ndata is ready to be imported into a standard installation of\nSparrow.\n\nIf connected to a Sparrow installation's PostgreSQL database,\nthe module can be used to insert data into the appropriate tables.\n\n## Key functions\n\n- `validate_data(schema_name: str, data: dict)`  \n  Checks data against a loader schema\n- `show_loader_schemas(schema_name: str,  ..., nest_depth=0)`  \n  Show the fields for one or several loader schemas.\n\n## Installation\n\n`pip install sparrow-loader`\n\nRequires Python `>=3.9`",
    'author': 'Daven Quinn',
    'author_email': 'dev@davenquinn.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from marshmallow import ValidationError
from pytest import raises
from . import show_loader_schemas, validate_data


def test_show_loader_schemas():
    show_loader_schemas()


def test_show_single_schema():
    show_loader_schemas("sample")


def test_validate_data():
    validate_data("sample", {"name": "test"})


def test_validate_sample_with_no_name():
    with raises(ValidationError):
        validate_data("sample", {"nameza": None})



from stringcase import pascalcase

# from marshmallow_sqlalchemy.fields import get_primary_keys
from macrostrat.database.mapper.utils import primary_key
from sqlalchemy.ext.automap import generate_relationship


def column_is_required(col):
    return all(
        [
            not col.nullable,  # Column cannot be null
            col.server_default is None,  # and has neither a database-side default
            col.default is None,  # nor an ORM default
        ]
    )


def to_schema_name(name):
    return pascalcase(name + "_schema")


def columns_for_prop(prop):
    try:
        return getattr(prop, "columns")
    except AttributeError:
        return list(getattr(prop, "local_columns"))


def prop_is_required(prop):
    cols = [column_is_required(c) for c in columns_for_prop(prop)]
    return any(cols)


def pk_values(instance):
    return primary_key(instance).values()


def pk_data(model, data):
    pk = primary_key(model)
    keys = {key: data.get(key) for key in pk.keys()}
    return keys.values()


def is_pk_defined(instance):
    vals = pk_values(instance)
    return all([v is not None for v in vals])


# Copied from sparrow.database.mapper, for now
def _gen_relationship(
    base, direction, return_fn, attrname, local_cls, referred_cls, **kw
):
    support_schemas = ["vocabulary", "core_view"]
    if (
        local_cls.__table__.schema in support_schemas
        and referred_cls.__table__.schema is None
    ):
        # Don't create relationships on vocabulary and core_view models back to the main schema
        return
    return generate_relationship(
        base, direction, return_fn, attrname, local_cls, referred_cls, **kw
    )
